﻿using System;
using System.Collections.Generic;
using System.Security;
using System.Text;
using System.Xml;
using System.Net;
using System.IO;
using System.Windows.Forms;
using Microsoft.Win32;
using System.Security.Cryptography;

namespace WordpressPlugin
{
    /// <summary>
    /// Helper class to interact with a blog service through the endpoint
    /// </summary>
    public static class XmlRpcHelper
    {
        /// <summary>
        /// Generic function for sending xml-rpc requests
        /// </summary>
        /// <param name="xml"></param>
        /// <param name="context"></param>
        /// <returns></returns>
        public static XmlDocument SendRequest(string xml, BlogContext context)
        {
            try
            {
                WebRequest webRequest = WebRequest.Create(context.PostApiUrl);
                webRequest.Method = "POST";
                
                // Send out the xml
                using (Stream request = webRequest.GetRequestStream())
                {
                    using (StreamWriter sw = new StreamWriter(request))
                    {
                        sw.Write(xml);
                    }
                }
                
                // Read in the server response
                string responseXml;
                using (WebResponse webResponse = webRequest.GetResponse())
                {
                    using (Stream response = webResponse.GetResponseStream())
                    {
                        using (StreamReader sr = new StreamReader(response))
                        {
                            responseXml = sr.ReadToEnd();
                        }
                    }
                }


                XmlDocument doc = new XmlDocument();
                doc.LoadXml(responseXml);

                // Look to see if there was an error, if so, read out the error codes and throw an exception
                XmlNodeList errorNodeList = doc.SelectNodes("//methodResponse/fault/value/struct/member");

                if (errorNodeList.Count > 0)
                {
                    XmlRpcException exception = new XmlRpcException();
                    foreach (XmlNode node in errorNodeList)
                    {
                        string name = GetInnerTextSafe(node.SelectSingleNode("./name"));
                        exception.Data[name] = GetInnerTextSafe(node.SelectSingleNode("./value"));
                    }

                    if (exception.ErrorCode == "403")
                    {
                        // If the there was some kind of permissions problem, clear out any saved passwords
                        XmlRpcHelper.ClearSavedPassword(context);
                    }

                    throw exception;
                }

                return doc;
            }
            catch (Exception we)
            {
                // Tell the user what the error was
                MessageBox.Show(we.Message, "Error while contacting blog service");
                throw new OperationCanceledException(we.Message, we);
            }
        }



        public static XmlDocument WpSetOptions(BlogContext context, string title, string tagline)
        {
            string xml = String.Format(XmlRequests.WpSetOptions, context.BlogId, SecurityElement.Escape(context.Username), SecurityElement.Escape(GetPassword(context)), SecurityElement.Escape(tagline), SecurityElement.Escape(title));
            return XmlRpcHelper.SendRequest(xml, context);
        }

        public static XmlDocument WpGetOptions(BlogContext context)
        {
            string xml = String.Format(XmlRequests.WpGetOptions, context.BlogId, SecurityElement.Escape(context.Username), SecurityElement.Escape(GetPassword(context)));

            return XmlRpcHelper.SendRequest(xml, context);
        }

        public static XmlDocument WpEditComment(BlogContext context, Comment c, CommentStatus status)
        {
            string xml = String.Format(XmlRequests.WpEditComment, status != CommentStatus.delete ? "wp.editComment" : "wp.deleteComment", context.BlogId, SecurityElement.Escape(context.Username), SecurityElement.Escape(GetPassword(context)), c.CommandId, status != CommentStatus.delete ? string.Format(XmlRequests.WpEditCommentParam, status) : "");

            return XmlRpcHelper.SendRequest(xml, context);
        }

        public static XmlDocument WpGetComments(BlogContext context, string status)
        {
            string xml = String.Format(XmlRequests.WpGetComments, context.BlogId, SecurityElement.Escape(context.Username), SecurityElement.Escape(GetPassword(context)), status);


            return XmlRpcHelper.SendRequest(xml, context);
        }

        public static XmlDocument MetaWeblogGetPost(BlogContext context)
        {
            string xml = String.Format(XmlRequests.MetaWeblogGetPost, context.PublishingContext.PostInfo.Id, SecurityElement.Escape(context.Username), SecurityElement.Escape(GetPassword(context)));

            return SendRequest(xml, context);
        }

        public static XmlDocument MetaWeblogEditPost(BlogContext context, string postXml)
        {
            string requestXml = String.Format(XmlRequests.MetaWeblogEditPost, context.PublishingContext.PostInfo.Id, SecurityElement.Escape(context.Username), SecurityElement.Escape(GetPassword(context)), postXml);
            return XmlRpcHelper.SendRequest(requestXml, context);
        }

        public static string GetPassword(BlogContext context)
        {
            // Look to see if we have a saved password in the registery already
            string password = GetSavedPassword(context);

            if (!string.IsNullOrEmpty(password))
                return password;

            // If we don't have a saved password, we need to prompt the user for one
            using (WeblogPasswordForm f = new WeblogPasswordForm(context))
            {
                if (f.ShowDialog() == DialogResult.OK)
                {
                    if (f.SavePassword)
                    {
                        SavePassword(context, f.Password);
                    }

                    // Save the password in memory for this session
                    InMemoryPassword[context.PublishingContext.AccountId] = f.Password;

                    return f.Password;
                }
            }

            // the user clicked 'cancel'
            throw new OperationCanceledException();
        }

        private static Dictionary<string, string> _memoryPassword = new Dictionary<string, string>();
        private static Dictionary<string, string> InMemoryPassword
        {
            get
            {
                return _memoryPassword;
            }
            set
            {
                _memoryPassword = value;
            }
        }

        private static readonly string WRITER_REG_PATH = "Software\\Microsoft\\WordpressPlugin\\";
        private static readonly byte[] CryptKey = new byte[] { 4, 3, 4, 5, 6, 5 };

        private static void SavePassword(BlogContext context, string password)
        {
            Registry.CurrentUser.CreateSubKey("Software\\Microsoft\\WordpressPlugin");

            password = Convert.ToBase64String(ProtectedData.Protect(Encoding.UTF8.GetBytes(password), CryptKey, DataProtectionScope.LocalMachine));

            Registry.CurrentUser.SetValue(WRITER_REG_PATH + context.PublishingContext.AccountId, password);
        }

        public static void ClearSavedPassword(BlogContext context)
        {
            InMemoryPassword.Remove(context.PublishingContext.AccountId);
            Registry.CurrentUser.DeleteValue(WRITER_REG_PATH + context.PublishingContext.AccountId, false);
        }

        public static string GetSavedPassword(BlogContext context)
        {
            // Look to see if there is a password saved in memory for this session
            string password;
            if (InMemoryPassword.TryGetValue(context.PublishingContext.AccountId, out password) && !string.IsNullOrEmpty(password))
                return password;

            // Look for a password in the registery, return null if we didnt find the password
            password = Registry.CurrentUser.GetValue(WRITER_REG_PATH + context.PublishingContext.AccountId, null) as string;

            if(string.IsNullOrEmpty(password))
                return null;

            try
            {
                byte[] encodedPlaintext = ProtectedData.Unprotect(Convert.FromBase64String(password), CryptKey, DataProtectionScope.LocalMachine); 
                return Encoding.UTF8.GetString(encodedPlaintext); 

            }
            catch(Exception)
            {
                return null;
            }
        }

        /// <summary>
        /// Helper function to make sure that if we use an xpath to find a node but dont find it, we 
        /// don't end up hitting a null ref
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        public static string GetInnerTextSafe(XmlNode node)
        {
            if (node == null)
                return String.Empty;

            return node.InnerText;
        }
    }

    public class XmlRpcException : Exception
    {
        public override string Message
        {
            get
            {
                return Data["faultString"] as string;
            }
        }

        public string ErrorCode
        {
            get
            {
                return Data["faultCode"] as string;
            }
        }
    }

}
